# To use this code, make sure you
#
#     import json
#
# and then, to convert JSON from a string, do
#
#     result = payload_from_dict(json.loads(json_string))
#     result = envelope_from_dict(json.loads(json_string))
#     result = payload_t_from_dict(json.loads(json_string))
#     result = purple_t_from_dict(json.loads(json_string))
#     result = envelope_t_from_dict(json.loads(json_string))
#     result = payload_t1_from_dict(json.loads(json_string))
#     result = t1_from_dict(json.loads(json_string))

from typing import Dict, Any, List, TypeVar, Callable, Type, cast
from enum import Enum


T = TypeVar("T")
EnumT = TypeVar("EnumT", bound=Enum)


def from_dict(f: Callable[[Any], T], x: Any) -> Dict[str, T]:
    assert isinstance(x, dict)
    return { k: f(v) for (k, v) in x.items() }


def from_str(x: Any) -> str:
    assert isinstance(x, str)
    return x


def from_list(f: Callable[[Any], T], x: Any) -> List[T]:
    assert isinstance(x, list)
    return [f(y) for y in x]


def from_float(x: Any) -> float:
    assert isinstance(x, (float, int)) and not isinstance(x, bool)
    return float(x)


def to_class(c: Type[T], x: Any) -> dict:
    assert isinstance(x, c)
    return cast(Any, x).to_dict()


def to_float(x: Any) -> float:
    assert isinstance(x, float)
    return x


def to_enum(c: Type[EnumT], x: Any) -> EnumT:
    assert isinstance(x, c)
    return x.value


class PayloadT:
    data: Dict[str, Any]
    kind: str

    def __init__(self, data: Dict[str, Any], kind: str) -> None:
        self.data = data
        self.kind = kind

    @staticmethod
    def from_dict(obj: Any) -> 'PayloadT':
        assert isinstance(obj, dict)
        data = from_dict(lambda x: x, obj.get("data"))
        kind = from_str(obj.get("kind"))
        return PayloadT(data, kind)

    def to_dict(self) -> dict:
        result: dict = {}
        result["data"] = from_dict(lambda x: x, self.data)
        result["kind"] = from_str(self.kind)
        return result


class PayloadT1:
    data: Dict[str, Any]
    kind: str

    def __init__(self, data: Dict[str, Any], kind: str) -> None:
        self.data = data
        self.kind = kind

    @staticmethod
    def from_dict(obj: Any) -> 'PayloadT1':
        assert isinstance(obj, dict)
        data = from_dict(lambda x: x, obj.get("data"))
        kind = from_str(obj.get("kind"))
        return PayloadT1(data, kind)

    def to_dict(self) -> dict:
        result: dict = {}
        result["data"] = from_dict(lambda x: x, self.data)
        result["kind"] = from_str(self.kind)
        return result


class V(Enum):
    A = "A"


class EnvelopeT:
    data: PayloadT1
    dst: List[str]
    id: str
    src: str
    t: float
    ttl: float
    v: V

    def __init__(self, data: PayloadT1, dst: List[str], id: str, src: str, t: float, ttl: float, v: V) -> None:
        self.data = data
        self.dst = dst
        self.id = id
        self.src = src
        self.t = t
        self.ttl = ttl
        self.v = v

    @staticmethod
    def from_dict(obj: Any) -> 'EnvelopeT':
        assert isinstance(obj, dict)
        data = PayloadT1.from_dict(obj.get("data"))
        dst = from_list(from_str, obj.get("dst"))
        id = from_str(obj.get("id"))
        src = from_str(obj.get("src"))
        t = from_float(obj.get("t"))
        ttl = from_float(obj.get("ttl"))
        v = V(obj.get("v"))
        return EnvelopeT(data, dst, id, src, t, ttl, v)

    def to_dict(self) -> dict:
        result: dict = {}
        result["data"] = to_class(PayloadT1, self.data)
        result["dst"] = from_list(from_str, self.dst)
        result["id"] = from_str(self.id)
        result["src"] = from_str(self.src)
        result["t"] = to_float(self.t)
        result["ttl"] = to_float(self.ttl)
        result["v"] = to_enum(V, self.v)
        return result


def payload_from_dict(s: Any) -> PayloadT:
    return PayloadT.from_dict(s)


def payload_to_dict(x: PayloadT) -> Any:
    return to_class(PayloadT, x)


def envelope_from_dict(s: Any) -> EnvelopeT:
    return EnvelopeT.from_dict(s)


def envelope_to_dict(x: EnvelopeT) -> Any:
    return to_class(EnvelopeT, x)


def payload_t_from_dict(s: Any) -> PayloadT:
    return PayloadT.from_dict(s)


def payload_t_to_dict(x: PayloadT) -> Any:
    return to_class(PayloadT, x)


def purple_t_from_dict(s: Any) -> Dict[str, Any]:
    return from_dict(lambda x: x, s)


def purple_t_to_dict(x: Dict[str, Any]) -> Any:
    return from_dict(lambda x: x, x)


def envelope_t_from_dict(s: Any) -> EnvelopeT:
    return EnvelopeT.from_dict(s)


def envelope_t_to_dict(x: EnvelopeT) -> Any:
    return to_class(EnvelopeT, x)


def payload_t1_from_dict(s: Any) -> PayloadT1:
    return PayloadT1.from_dict(s)


def payload_t1_to_dict(x: PayloadT1) -> Any:
    return to_class(PayloadT1, x)


def t1_from_dict(s: Any) -> Dict[str, Any]:
    return from_dict(lambda x: x, s)


def t1_to_dict(x: Dict[str, Any]) -> Any:
    return from_dict(lambda x: x, x)



from datetime import datetime, timezone, tzinfo
import hashlib
import json
import unittest
import unittest.mock

from .lang.python.envelope import EnvelopeT, PayloadT

from .simple_envelope import HashCollector, JsonCollector, JsonProps, SimpleEnvelopeProps, simpleEnvelope, sortKeys


class Mockdatetime:
    def now(self):
        return datetime.fromtimestamp(1624140000)

mockdatetime = Mockdatetime()

def toSVals(calls: list[unittest.mock.call]):
    return list(map(lambda x: list(map(lambda x: x.to_dict(), x.args)), calls))

class SimpleEnvelopeTest(unittest.TestCase):

    def test_simple_hash(self):
        hashC = HashCollector()
        sortKeys({
            'kind': "test",
            'data': {
                'name': "object",
                'date': "2021-05-20",
            },
        },
            lambda sval: hashC.append(sval))
        self.assertEqual(
            hashC.digest(), "5zWhdtvKuGob1FbW9vUGPQKobcLtYYr5wU8AxQRVraeB")

    def test_serialization(self):
        self.maxDiff = None
        message: SimpleEnvelopeProps = SimpleEnvelopeProps(**{
            'src': "test case",
            'data': PayloadT.from_dict({
                'kind': "test",
                'data': {'name': "object", 'date': "2021-05-20"}
            }),
            'id': "1624140000000-4a2a6fb97b3afe6a7ca4c13457c441664c7f6a6c2ea7782e1f2dea384cf97cb8",
            'dst': [],
            't': datetime.fromtimestamp(0.444, tz=timezone.utc),
            'ttl': 10
        })

        stringify = '{"data":{"data":{"date":"2021-05-20","name":"object"},"kind":"test"},"dst":[],"id":"1624140000000-4a2a6fb97b3afe6a7ca4c13457c441664c7f6a6c2ea7782e1f2dea384cf97cb8","src":"test case","t":444,"ttl":10,"v":"A"}'
        # print(f'---:{stringify}')
        jsonStr = simpleEnvelope(message).asJson()
        # print(f'===:{jsonStr}')
        self.assertEqual(jsonStr, stringify)

    def test_serialization_with_hash(self):
        self.maxDiff = None
        message: SimpleEnvelopeProps = SimpleEnvelopeProps(**{
            'src': "test case",
            'data': PayloadT.from_dict({
                'kind': "test",
                'data': {'name': "object", 'date': "2021-05-20"},
            }),
            'dst': [],
            'ttl': 10,
            'datetime': mockdatetime
        })
        stringify = '{"data":{"data":{"date":"2021-05-20","name":"object"},"kind":"test"},"dst":[],"id":"1624140000000-BbYxQMurpUmj1W6E4EwYM79Rm3quSz1wwtNZDSsFt1bp\","src":"test case","t":1624140000000,"ttl":10,"v":"A"}'
        self.assertEqual(simpleEnvelope(message).asJson(), stringify)

    def test_serialization_with_intend(self):
        self.maxDiff = None
        message: SimpleEnvelopeProps = SimpleEnvelopeProps(**{
            'src': "test case",
            'data': PayloadT.from_dict({
                'kind': "test",
                'data': {'name': "object", 'date': "2021-05-20"},
            }),
            'dst': [],
            'ttl': 10,
            'jsonProp': JsonProps(**{
                'indent': 2,
            }),
            'datetime': mockdatetime
        })
        stringify = json.dumps(
            json.loads(
                '{"data":{"data":{"date":"2021-05-20","name":"object"},"kind":"test"},"dst":[],"id":"1624140000000-BbYxQMurpUmj1W6E4EwYM79Rm3quSz1wwtNZDSsFt1bp","src":"test case","t":1624140000000,"ttl":10,"v":"A"}'
            ),
            indent=2)
        self.assertEqual(simpleEnvelope(message).asJson(), stringify)

    def test_sort_with_out_with_string(self):
        fn = unittest.mock.Mock()
        sortKeys("string", fn)
        self.assertEqual(toSVals(fn.mock_calls), [[{'val': {'val': "string"}}]])

    def test_sort_with_out_with_date(self):
        fn = unittest.mock.Mock()
        sortKeys(datetime.fromtimestamp(0.444, tz=timezone.utc), fn)
        self.assertEqual(toSVals(fn.mock_calls), 
            [[{'val': {'val': datetime.fromtimestamp(0.444, tz=timezone.utc)}}]])

    def test_sort_with_out_with_number(self):
        fn = unittest.mock.Mock()
        sortKeys(4711, fn)
        self.assertEqual(toSVals(fn.mock_calls), [[{'val': {'val': 4711}}]])

    def test_sort_with_out_with_boolean(self):
        fn = unittest.mock.Mock()
        sortKeys(False, fn)
        self.assertEqual(toSVals(fn.mock_calls), [[{'val': {'val': False}}]])

    def test_sort_with_out_with_array_of_empty(self):
        fn = unittest.mock.Mock()
        sortKeys([], fn)
        self.assertEqual(toSVals(fn.mock_calls), 
            [[{'outState': "AS"}], [{'outState': "AE"}]])

    def test_sort_with_out_with_array_of_1_2(self):
        fn = unittest.mock.Mock()
        sortKeys([1, 2], fn)
        self.assertEqual(toSVals(fn.mock_calls), [
            [{'outState': "AS"}],
            [{'val': {'val': 1}}],
            [{'val': {'val': 2}}],
            [{'outState': "AE"}],
        ])

    def test_sort_with_out_with_array_of_1_2_3_4(self):
        fn = unittest.mock.Mock()
        sortKeys(
            [
                [1, 2],
                [3, 4],
            ],
            fn
        )
        self.assertEqual(toSVals(fn.mock_calls), [
            [{'outState': "AS"}],
            [{'outState': "AS"}],
            [{'val': {'val': 1}}],
            [{'val': {'val': 2}}],
            [{'outState': "AE"}],
            [{'outState': "AS"}],
            [{'val': {'val': 3}}],
            [{'val': {'val': 4}}],
            [{'outState': "AE"}],
            [{'outState': "AE"}],
        ])

    def test_sort_with_out_with_obj_of_empty_obj(self):
        fn = unittest.mock.Mock()
        sortKeys({}, fn)
        self.assertEqual(toSVals(fn.mock_calls), 
            [[{'outState': "OS"}], [{'outState': "OE"}]])

    def test_sort_with_out_with_obj_of_obj_y_1_x_2(self):
        fn = unittest.mock.Mock()
        sortKeys({'y': 1, 'x': 2}, fn)
        self.assertEqual(toSVals(fn.mock_calls), [
            [{'outState': "OS"}],
            [{'attribute': "x"}],
            [{'val': {'val': 2}}],
            [{'attribute': "y"}],
            [{'val': {'val': 1}}],
            [{'outState': "OE"}],
        ])

    def test_sort_with_out_with_obj_of_obj_y_b_1_a_2(self):
        fn = unittest.mock.Mock()
        sortKeys({'y': {'b': 1, 'a': 2}}, fn)
        self.assertEqual(toSVals(fn.mock_calls), [
            [{'outState': "OS"}],
            [{'attribute': "y"}],
            [{'outState': "OS"}],
            [{'attribute': "a"}],
            [{'val': {'val': 2}}],
            [{'attribute': "b"}],
            [{'val': {'val': 1}}],
            [{'outState': "OE"}],
            [{'outState': "OE"}],
        ])

    def test_JSONCollector_empty_obj(self):
        out = []
        json = JsonCollector(lambda o: out.append(o))
        sortKeys({}, lambda o: json.append(o))
        self.assertEqual("".join(out), "{}")

    def test_JSONCollector_empty_array(self):
        out = []
        json = JsonCollector(lambda o: out.append(o))
        sortKeys([], lambda o: json.append(o))
        self.assertEqual("".join(out), "[]")

    def test_JSONCollector_x_y_1_z_x_y_z(self):
        out = []
        json = JsonCollector(lambda o: out.append(o))
        sortKeys({'x': {'y': 1, 'z': "x"}, 'y': {}, 'z': []},
                 lambda o: json.append(o))
        self.assertEqual("".join(out), '{"x":{"y":1,"z":"x"},"y":{},"z":[]}')

    def test_JSONCollector_array_xx(self):
        out = []
        json = JsonCollector(lambda o: out.append(o))
        sortKeys(["xx"], lambda o: json.append(o))
        self.assertEqual("".join(out), '["xx"]')

    def test_JSONCollector_array_1_2(self):
        out = []
        json = JsonCollector(lambda o: out.append(o))
        sortKeys([1, "2"], lambda o: json.append(o))
        self.assertEqual("".join(out), '[1,"2"]')

    def test_JSONCollector_1_2_A(self):
        out = []
        json = JsonCollector(lambda o: out.append(o))
        sortKeys([1, ["2", "A"], "E"], lambda o: json.append(o))
        self.assertEqual("".join(out), '[1,["2","A"],"E"]')

    def test_JSONCollector_indent_2_empty_obj(self):
        out = []
        json = JsonCollector(lambda o: out.append(o), JsonProps(**{'indent': 2}))
        sortKeys({}, lambda o: json.append(o))
        self.assertEqual("".join(out), "{}")

    def test_JSONCollector_indent_2_array_empty(self):
        out = []
        json = JsonCollector(lambda o: out.append(o), JsonProps(**{'indent': 2}))
        sortKeys([], lambda o: json.append(o))
        self.assertEqual("".join(out), "[]")

    def test_JSONCollector_indent_2_x_y_1_z_x(self):
        out = []
        json = JsonCollector(lambda o: out.append(o), JsonProps(**{'indent': 2}))
        sortKeys({'x': {'y': 1, 'z': "x"}, 'y': {}, 'z': []},
                 lambda o: json.append(o))
        self.assertEqual(
            "".join(out), '{\n  "x": {\n    "y": 1,\n    "z": "x"\n  },\n  "y": {},\n  "z": []\n}')

    def test_JSONCollector_indent_2_xx(self):
        out = []
        json = JsonCollector(lambda o: out.append(o), JsonProps(**{'indent': 2}))
        sortKeys(["xx"], lambda o: json.append(o))
        self.assertEqual("".join(out), '[\n  "xx"\n]')

    def test_JSONCollector_indent_2_array_1_2(self):
        out = []
        json = JsonCollector(lambda o: out.append(o), JsonProps(**{'indent': 2}))
        sortKeys([1, "2"], lambda o: json.append(o))
        self.assertEqual("".join(out), '[\n  1,\n  "2"\n]')

    def test_JSONCollector_1_date444(self):
        out = []
        json = JsonCollector(lambda o: out.append(o))
        obj = [1, datetime.fromtimestamp(0.444, tz=timezone.utc)]
        sortKeys(obj, lambda o: json.append(o))
        self.assertEqual("[1,\"1970-01-01T00:00:00.444Z\"]", "".join(out))

    def test_HashCollector_1(self):
        hash = HashCollector()
        sortKeys({'x': {'y': 1, 'z': "x"}, 'y': {}, 'z': [],
                 'd': datetime.fromtimestamp(0.444, tz=timezone.utc)}, lambda o: hash.append(o))
        self.assertEqual(
            hash.digest(), "5PvJAWGkaKAHax6tsaKGfPYm6JfXxZs15wRTDpSKaZ2G")

    def test_HashCollector_2(self):
        hash = HashCollector()
        sortKeys({'x': {'y': 2, 'z': "x"}, 'y': {}, 'z': [],
                 'date': datetime.fromtimestamp(0.444, tz=timezone.utc)}, lambda o: hash.append(o))
        self.assertEqual(
            hash.digest(), "ECVWfmcNaUGkgvPZe7CojrnRNULxNczKXU8PGns6UDvr")

    def test_HashCollector_3(self):
        hash = HashCollector()
        sortKeys({'x': {'x': 1, 'z': "x"}, 'y': {}, 'z': [],
                 'date': datetime.fromtimestamp(0.444, tz=timezone.utc)}, lambda o: hash.append(o))
        self.assertEqual(
            hash.digest(), "EoYNGMtap1k9iEAGeVtHmJwpMjQLKWJmR27SG6aC9fSg")

    def test_HashCollector_4(self):
        hash1 = HashCollector()
        sortKeys({'x': {'x': 1, 'z': "x"}, 'y': {}, 'z': [],
                 'date': datetime.fromtimestamp(0.444, tz=timezone.utc)}, lambda o: hash1.append(o))
        hash2 = HashCollector()
        sortKeys({'date': datetime.fromtimestamp(0.444, tz=timezone.utc), 'x': {
                 'x': 1, 'z': "x"}, 'y': {}, 'z': []}, lambda o: hash2.append(o))
        self.assertEqual(hash1.digest(), hash2.digest())

    def test_HashCollector_3_internal_update(self):
        class MockHash:
            hash: any
            mockUpdate: any
            def __init__(self) -> None:
                self.mockUpdate = unittest.mock.Mock()
                self.hash = hashlib.new('sha256')

            def update(self, a: any):
                self.mockUpdate(a) 
                self.hash.update(a)

            def digest(self):
                return self.hash.digest()

        hashCollector = HashCollector(MockHash())
        sortKeys({'x': {'r': 1, 'z': "u"}, 'y': {}, 'z': [], 'date': datetime.fromtimestamp(
            0.444, tz=timezone.utc)}, lambda o: hashCollector.append(o))
        result = list(map(lambda m: m.args[0].decode(), hashCollector.hash.mockUpdate.mock_calls))
        self.assertEqual(result, 
            ["date", "1970-01-01T00:00:00.444Z", "x", "r", "1", "z", "u", "y", "z"])
        self.assertEqual(hashCollector.digest(),
                         "CwEMjUHV6BpDS7AGBAYqjY6qMKE6xC8Z56H5T2ZuUuXe")

    def test_missing_data_in_envelope(self):
        payt: PayloadT = PayloadT.from_dict({
            'kind': 'kind',
            'data': { 'y': 4 }
        })
        message: SimpleEnvelopeProps = SimpleEnvelopeProps(**{
            'src': "test case",
            'data': payt
        })
        ref = EnvelopeT.from_dict(json.loads(simpleEnvelope(message).asJson()))
        self.assertEqual(simpleEnvelope(SimpleEnvelopeProps(**ref.to_dict())).asEnvelope().data.to_dict(), payt.to_dict())


if __name__ == '__main__':
    unittest.main()




from dataclasses import dataclass

import typing
# import unittest
import unittest.mock

from .lang.python.envelope import *
import json
from datetime import datetime
import hashlib
import math
from base58 import b58encode


class JsonProps:
    indent: int
    newLine: str

    def __init__(self, indent = None, newLine = None) -> None:
        self.indent = 0 if indent is None else indent
        self.newLine = "\n" if newLine is None else newLine

class SimpleEnvelopeProps:
    id: str
    src: str
    dst: str
    t: datetime
    ttl: int
    data: PayloadT
    jsonProp: JsonProps
    datetime: datetime

    def __init__(self, src, data, id = None, dst = [], t = None, ttl = 10, jsonProp = JsonProps(), datetime = datetime, v = None) -> None:
        if isinstance(data, PayloadT):
            self.data = data
        else:
            self.data = PayloadT.from_dict(data)
        self.src = src
        self.id = id
        self.dst = dst
        self.t = t
        self.ttl = ttl
        self.jsonProp = jsonProp
        self.datetime = datetime

# type OutputFn = (str: string) => void;

# tyype ValueType = string | number | boolean | Date | undefined;


class ValType:
    def toString(): str
    def asValue(): any
    def to_dict(): dict

def jsIsoFormat(val: datetime):
    isoStr = val.isoformat().split(".")
    return f'{isoStr[0]}.{isoStr[1][0:3]}Z'

class JsonValType(ValType):
    val: any

    def __init__(self, val: any):
        self.val = val

    def asValue(self):
        return self.val

    def toString(self):
        val = self.val
        if isinstance(self.val, float): 
           if float(self.val) == int(self.val):
            val = int(self.val)
        elif isinstance(self.val, datetime):
            val = jsIsoFormat(self.val)
        return json.dumps(val)
        # except Exception as e:
        # print("XXXXXXXX[", self.val, e)
    def to_dict(self):
        return {
            'val': self.val
        }

class PlainValType(ValType):
    val: str

    def __init__(self, val: str):
        self.val = val

    def asValue(self):
        return self.val

    def toString(self):
        return self.val

    def to_dict(self):
        return {
            'val': self.val
        }


class OutState(Enum):
    NONE = "NE",
    ARRAY_START = "AS",
    ARRAY_END = "AE",
    OBJECT_START = "OS",
    OBJECT_END = "OE",


class SVal:
    attribute: str
    val: any
    outState: OutState

    def __init__(self, attribute = None, val = None, outState = None) -> None:
        self.attribute = attribute
        self.val = val
        self.outState = outState

    def to_dict(self):
        ret = {}
        if self.attribute is not None:
            ret['attribute'] = self.attribute
        if self.val is not None:
            ret['val'] = self.val.to_dict()
        if self.outState is not None:
            ret['outState'] = self.outState.value[0]
        return ret

OutputFN = typing.Callable[[str], None]

class JsonCollector:
    output: OutputFN
    indent: str
    commas: list[str]
    elements: list[int]
    props: JsonProps
    nextLine: str
    attribute: str

    def __init__(self, output: OutputFN, props: JsonProps = JsonProps()):
        self.output = output
        self.props = props
        self.indent = (" " * self.props.indent)
        self.nextLine = self.props.newLine if self.props.indent > 0 else ""
        self.commas = [""]
        self.elements = [0]
        self.attribute = ""
        # print("JsonCollector::__init__")

    def suffix(self) -> str:
        if self.elements[-1] > 0:
            return self.nextLine + (self.indent * (len(self.commas) - 1))
        else:
            return ""

    def append(self, sval: SVal):
        # print(f"append:{sval.to_dict()}-{this.commas}")
        if sval.outState is not None:
            if sval.outState == OutState.ARRAY_START:
                # print(f"Array-Start:{this.commas}-{this.suffix()}-{this.attribute}")
                self.output(
                    self.commas[-1] +
                    self.suffix() +
                    (self.attribute if self.attribute else "") +
                    "["
                )
                self.attribute = None
                self.commas[-1] = ","
                self.commas.append("")
                self.elements.append(0)
                return
            if sval.outState == OutState.ARRAY_END:
                self.commas.pop()
                self.output(self.suffix() + "]")
                self.elements.pop()
                return
            if sval.outState == OutState.OBJECT_START:
                self.output(
                    self.commas[-1] +
                    self.suffix() +
                    (self.attribute if self.attribute is not None else "") +
                    "{"
                )
                self.attribute = None
                self.commas[-1] = ","
                self.commas.append("")
                self.elements.append(0)
                return
            if sval.outState == OutState.OBJECT_END:
                self.commas.pop()
                self.output(self.suffix() + "}")
                self.elements.pop()
                return

        if sval.val is not None:
            self.elements[-1] = self.elements[-1] + 1
            # print(f"---[{sval.val}]-[{this.commas[-1]}]suffix[{this.suffix()}]attribute[{this.attribute}]val[{sval.val.toString()}]")
            out = self.commas[-1] + self.suffix() + (
                self.attribute if self.attribute is not None else "") + sval.val.toString()
            self.output(out)
            self.attribute = None
            self.commas[-1] = ","
        if (sval.attribute):
            self.elements[-1] = self.elements[-1] + 1
            self.attribute = json.dumps(
                sval.attribute) + ":" + (" " if len(self.indent) > 0 else "")


class HashCollector:
    # readonly hash: crypto.Hash = crypto.createHash("sha256");
    hash: any # hashlib._Hash

    def __init__(self, hash = None) -> None:
        self.hash = hashlib.new('sha256') if hash is None else hash

    def digest(self):
        return b58encode(self.hash.digest()).decode()

    def append(self, sval: SVal):
        if sval.outState is not None:
            return
        if sval.attribute is not None:
            tmp = sval.attribute.encode('utf-8')
            # print("attribute=", tmp)
            self.hash.update(tmp)
        if sval.val is not None:
            out = sval.val.asValue()
            if isinstance(out, datetime):
                out = jsIsoFormat(out)
            else:
                out = str(out)
            # print("val=", out)
            self.hash.update(out.encode("utf-8"))


def sortKeys(e: any, out: Callable[[SVal], None]):
    if isinstance(e, list):
        out(SVal(**{'outState': OutState.ARRAY_START}))
        for i in e:
            sortKeys(i, out)
        out(SVal(**{'outState': OutState.ARRAY_END}))
        return
    elif isinstance(e, dict):
        out(SVal(**{'outState': OutState.OBJECT_START}))
        keys = list(e.keys())
        keys.sort()
        # print("keys=", keys)
        for i in keys:
            out(SVal(**{'attribute': i}))
            sortKeys(e[i], out)

        out(SVal(**{'outState': OutState.OBJECT_END}))
        return
    else:
        # print(f"VAL[{e}]")
        out(SVal(**{'val': JsonValType(e)}))
        return


@dataclass
class JsonHash:
    jsonStr: str
    hash: str


# @unittest.mock.patch("datetime.now")
class SimpleEnvelope:
    simpleEnvelopeProps: SimpleEnvelopeProps
    envJsonStrings: list[str]
    envJsonC: JsonCollector
    envelope: EnvelopeT
    dataJsonHash: JsonHash
    datetime: datetime

    def __init__(self, env: SimpleEnvelopeProps):
        self.envJsonStrings = []
        self.simpleEnvelopeProps = env
        # self.simpleEnvelopeProps.jsonProp = JsonProps() if self.simpleEnvelopeProps.jsonProp is None else self.simpleEnvelopeProps.jsonProp
        # print(this.simpleEnvelopeProps.jsonProp)
        self.envJsonC = JsonCollector(
            lambda part: self.envJsonStrings.append(part),
            self.simpleEnvelopeProps.jsonProp
        )
        self.datetime = env.datetime if env.datetime is not None else datetime

    def asDataJson(self):
        return self.dataJsonHash.jsonStr

    def asJson(self):
        str = "".join(self.lazy().envJsonStrings)
        # Caution
        self.asJson = lambda: str
        return str

    def asEnvelope(self):
        # Caution
        self.asEnvelope = lambda: self.envelope
        return self.lazy().envelope

    def toDataJson(self) -> JsonHash:
        dataJsonStrings: list[str] = []
        dataJsonC = JsonCollector(lambda part: dataJsonStrings.append(part), JsonProps(**{
            'indent': self.simpleEnvelopeProps.jsonProp.indent,
            'newLine': "\n" + (" " * (self.simpleEnvelopeProps.jsonProp.indent * 2))
        }))
        dataHashC: HashCollector = None
        dataProcessor = None
        if self.simpleEnvelopeProps.id is not None:
            def dataProcessor(sval): return dataJsonC.append(sval)
        else:
            dataHashC = HashCollector()

            def my(sval: SVal):
                dataHashC.append(sval)
                dataJsonC.append(sval)
            dataProcessor = my

        sortKeys(self.simpleEnvelopeProps.data.data, dataProcessor)
        return JsonHash(**{
            'jsonStr': "".join(dataJsonStrings),
            'hash': dataHashC.digest() if dataHashC is not None else None
        })

    def lazy(self):
        self.dataJsonHash = self.toDataJson()
        t = 0
        if isinstance(self.simpleEnvelopeProps.t, datetime):
            date = self.simpleEnvelopeProps.t if self.simpleEnvelopeProps.t is not None else self.datetime.now()
            t = int(date.timestamp()*1000)
        elif isinstance(self.simpleEnvelopeProps.t, float) or isinstance(self.simpleEnvelopeProps.t, int):
            t = self.simpleEnvelopeProps.t
        else:
            t = int(self.datetime.now().timestamp()*1000)
        # print(">>>>>>", self.simpleEnvelopeProps.t.__class__)
        envelope: EnvelopeT = EnvelopeT.from_dict({
            'v': "A",
            'id': self.simpleEnvelopeProps.id if self.simpleEnvelopeProps.id is not None else f'{t}-{self.dataJsonHash.hash}',
            'src': self.simpleEnvelopeProps.src,
            'dst': self.simpleEnvelopeProps.dst if self.simpleEnvelopeProps.dst is not None else [],
            't': t,
            'ttl': self.simpleEnvelopeProps.ttl if self.simpleEnvelopeProps.ttl is not None else 10,
            'data': PayloadT.from_dict({
                'kind': self.simpleEnvelopeProps.data.kind,
                'data': {}
            }).to_dict()
        }).to_dict()
        envelope['data']['data'] = None
        class Processor:
            nextValue: bool
            data: str
            envJsonC: JsonCollector
            def __init__(self, data: str, envJsonC: JsonCollector) -> None:
                self.data = data
                self.nextValue = False
                self.envJsonC = envJsonC
            def append(self, sval: SVal):
                oval = sval
                if sval.attribute == "data":
                    self.nextValue = True
                elif self.nextValue:
                    if sval.val is not None and sval.val.asValue() == None:
                        oval = SVal(**{
                            'val': PlainValType(self.data),
                        })
                    self.nextValue = False
                self.envJsonC.append(oval)

        processor = Processor(self.dataJsonHash.jsonStr, self.envJsonC)
        sortKeys(envelope, lambda sval: processor.append(sval))

        self.envelope = EnvelopeT.from_dict({
            **envelope,
            'data': {
                **envelope['data'],
                # 'kind': envelope.data.kind,
                'data': self.simpleEnvelopeProps.data.data
            },
        })
        self.lazy = lambda: self
        return self


def simpleEnvelope(env: SimpleEnvelopeProps) -> SimpleEnvelope:
    return SimpleEnvelope(env)



import os
import sys
import setuptools
import json
from shutil import copyfile
from setuptools.command.install import install
from pathlib import Path


def read(fname):
  with open(os.path.join(os.path.dirname(__file__), fname), 'rt') as f:
    return f.read()


def write_init(version: str):
  with open("./src/lang/python/version.py", "w") as init_file:
    print("__version__ = '{version}'".format(version=version), file=init_file)
  with open("./src/__init__.py", "w") as init_file:
    print("# generated", file=init_file)
    print("from .lang.python.envelope import *", file=init_file)
    print("from .simple_envelope import *", file=init_file)
    print("from .lang.python.version import __version__", file=init_file)
    print("if __name__ == '__main__':", file=init_file)
    print(" print('ready to format a c5 envelope')", file=init_file)

#class PreInstallCommand(install):
#    """Pre-installation for installation mode."""
#    def run(self):



#print(sys.argv)
#print(os.environ)

main_ns = {
        '__name__': 'hack'
        }

if Path("./package.json").is_file():
    version = json.loads(read('package.json'))['version']
    main_ns['__version__'] = version


#ver_path = convert_path('__init__.py')

if sys.argv[0] == 'setup.py' and sys.argv[1] == 'sdist':
    version = json.loads(read('package.json'))['version']
    write_init(version)
    #copyfile('./src/simple_envelope.py', './lang/python/simple_envelope.py')
    #copyfile('./package.json', './lang/python/package.json')
    #main_ns['__version__'] = version


lang_python_init = Path("./src/lang/python/version.py")
if lang_python_init.is_file():
    with open(lang_python_init) as ver_file:
      c = ver_file.read()
      exec(c, main_ns)

lang_python_init = Path("./lang/python/version.py")
if lang_python_init.is_file():
    with open(lang_python_init) as ver_file:
      c = ver_file.read()
      exec(c, main_ns)

#install.run(self)

if main_ns['__version__'] == 'hack':
    print("i-3424ri0jrejoifeowjofgjeajofewoifjoiweafgjoaogsgojavfds");

setuptools.setup(
  name='c5-envelope',
  version=main_ns['__version__'],
  author='Meno Abels',
  author_email='meno.abels@adviser.com',
  setup_requires=[],
  install_requires=[
      'base58'
  ],
  ext_modules=[],
  packages=['c5_envelope', 'c5_envelope.lang.python'],
  package_dir={
      'c5_envelope': 'src', 
      'c5_envelope.lang.python': 'src/lang/python'
  },

  description="C5-ENVELOPE Repository",
  long_description=read('README.md'),
  long_description_content_type="text/markdown",
  keywords = "data ocean lake",
  url = "https://github.com/mabels/envelope",
  classifiers=[
    "Intended Audience :: Developers",
    "Development Status :: 5 - Production/Stable",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Topic :: Scientific/Engineering",
    "Intended Audience :: Developers"
  ], 
)
